#include "stdafx.h"
#include "window.h"

#include "style.h"
#include "windows.h"
#include "state.h"

namespace GDIView
{

static const Style	BUTTON(0xff6f6fff, 0xff9f9fff, 0x3f0000ff),
					RADIO_ACTIVE(0xff00ff00, 0xfff9ff9f, 0x3f00ff00),
					RADIO_INACTIVE(0xffff3f3f, 0xffff9f9f, 0x3fff0000),
					DRAG(0xffff7f00, 0xffffff00, 0x3f3f1f00),
					DROP(0xff8f8f8f, 0xffafafaf, 0x3fffffff),
					SCROLL(0xff8f3f8f, 0xffaf3faf, 0x3f3f3f3f),
					WINDOW(0xff3f3f3f, 0xff5f5f5f, 0x3f7f7f7f),
					FOCUS(0xff4f4f4f, 0xff6f6f6f, 0x3f8f8f8f);

//local methods
const WCHAR* multiByteToWide(const char* aString)
{
	static WCHAR wide[MAX_PATH];

	MultiByteToWideChar(CP_ACP, 0, aString, (int)strlen(aString), wide, MAX_PATH);
	wide[strlen(aString)] = 0;

	return wide;
}

//implementation
Window::Window(const Windows& someWindows, State& aState, const WCHAR* aTitle, const int aX, const int aY, const int aWidth, const int aHeight)
:myWindows(someWindows)
,myState(aState)
,myTitle(aTitle)
,myX(aX)
,myY(aY)
,myWidth(aWidth)
,myHeight(aHeight)
,myNumRects(0)
,myNumTexts(0)
,myNumHandles(0)
,myWantsFocus(false)
{
}

void Window::rect(const int aX, const int aY, const int aWidth, const int aHeight)
{
	rect(DROP, aX, aY, aWidth, aHeight, false);
}

void Window::text(const int aX, const int aY, const WCHAR* aText, const unsigned int aColor)
{
	if(myNumTexts < MAX_TEXTS)
	{
		myTexts[myNumTexts].set(aX, aY + (myState.myButtonHeight - Text::FONT_SIZE) / 2 - 1, aText, aColor);
		myNumTexts++;
	}
}

const bool Window::button(const int aX, const int aY, const WCHAR* aText)
{
	if(aText)
		text(aX, aY, aText, 0xff000000);
	return rect(BUTTON, aX, aY, myState.myButtonWidth, myState.myButtonHeight, true) && wasPressed(VK_LBUTTON);
}

const bool Window::wideButton(const int aX, const int aY, const int aWidth, const WCHAR* aText)
{
	if(aText)
		text(aX, aY, aText, 0xff000000);
	return rect(BUTTON, aX, aY, aWidth, myState.myButtonHeight, true) && wasPressed(VK_LBUTTON);
}

const bool Window::radio(const bool anActive, const int aX, const int aY, const WCHAR* aText)
{
	if(aText)
		text(aX, aY, aText, 0xff000000);
	return rect(anActive ? RADIO_ACTIVE : RADIO_INACTIVE, aX, aY, myState.myButtonWidth, myState.myButtonHeight, true) && wasPressed(VK_LBUTTON);
}

const bool Window::drag(const void* aHandle, const int aX, const int aY, const int aWidth, const int aHeight, const WCHAR* aText)
{
	if(aHandle == myState.myDrag)
	{
		//do drag
		myState.doDrag(DRAG, aWidth, aHeight, aText);

		//being dragged
		return true;
	}
	else
	{
		if(aText)
			text(aX, aY, aText, 0xff000000);

		//does NOT require focus, hence the direct access of State::isDown()
		if(rect(DRAG, aX, aY, aWidth, aHeight, true) && !myState.myDrag && myState.isDown(VK_LBUTTON))
		{
			//setup drag
			myState.myDrag = aHandle;
			myState.beginDrag(aX + myX, aY + myY, DRAG, aWidth, aHeight, aText);

			//being dragged
			return true;
		}
	}

	//not being dragged
	return false;
}

const void* Window::drop(const int aX, const int aY, const int aWidth, const int aHeight)
{
	if(rect(DROP, aX, aY, aWidth, aHeight, true) && myState.myDrop)
		return myState.myDrop;
	else
		return NULL;
}

void Window::hScroll(const float aMax, const float aRange, float& aValue, const int aX, const int aY, const int aWidth, const int aHeight)
{
	const int SIZE((int)(aWidth * aRange / aMax));
	WString s;

	if(&aValue == myState.myScroll)
	{
		aValue += myState.scrollX() / (float)(aWidth - SIZE) * aMax;
		if(aValue < 0.f)
			aValue = 0.f;
		else if(aValue > aMax)
			aValue = aMax;
	}

	s.format(L"%.1f", aValue);
	text(aX + aWidth / 2 , aY, s, 0xff000000);

	//background
	rect(SCROLL, aX, aY, aWidth, aHeight, true);

	//dragabble thing
	if(rect(DRAG, aX + (int)((aWidth - SIZE) * aValue / aMax), aY, SIZE, aHeight, true) && !myState.myScroll && isDown(VK_LBUTTON))
		myState.myScroll = &aValue;
}

void Window::vScroll(const float aMax, const float aRange, float& aValue, const int aX, const int aY, const int aWidth, const int aHeight)
{
	const int SIZE((int)(aHeight * aRange / aMax));
	WString s;

	if(&aValue == myState.myScroll)
	{
		aValue += myState.scrollY() / (float)(aHeight - SIZE) * aMax;
		if(aValue < 0.f)
			aValue = 0.f;
		else if(aValue > aMax)
			aValue = aMax;
	}

	s.format(L"%.1f", aValue);
	text(aX, aY + aHeight / 2, s, 0xff000000);

	//background
	rect(SCROLL, aX, aY, aWidth, aHeight, true);

	//draggable thing
	if(rect(DRAG, aX, aY + (int)((aHeight - SIZE) * aValue / aMax), aWidth, SIZE, true) && !myState.myScroll && isDown(VK_LBUTTON))
		myState.myScroll = &aValue;
}

const bool Window::node(const void* aHandle, const int aX, const int aY, const WCHAR* aLabel)
{
	if(myNumHandles < MAX_HANDLES)
	{
		bool& h(handleState(aHandle));
		WString s;

		s.format(h ? L"-%s" : L"+%s", aLabel);

		if(radio(h, aX, aY, s))
			h = !h;

		return h;
	}

	return false;
}

void Window::stringCombo(unsigned int& aChoice, const int aX, const int aY, const WCHAR** someChoices)
{
	if(myNumHandles < MAX_HANDLES)
	{
		bool& h(handleState(&aChoice));

		//expanded
		if(h)
		{
			//current choice
			if(button(aX, aY, someChoices[aChoice]))
				h = false;	//same choice

			//list
			unsigned int c(0);
			int y(aY);
			while(someChoices[c])
			{
				if(radio(c == aChoice, aX, y += buttonHeight(), someChoices[c]))
				{
					aChoice = c;
					h = false;
				}
				c++;
			}
		}
		//collapsed
		else
		{
			if(radio(h, aX, aY, someChoices[aChoice]))
				h = true;
		}
	}
}

const int Window::buttonWidth() const
{
	return myState.myButtonWidth;
}

const int Window::buttonHeight() const
{
	return myState.myButtonHeight;
}

void Window::render(Gdiplus::Graphics& someGraphics)
{
	unsigned int i;

	//rects
	if(myNumRects)
	{
		for(i = 0; i < myNumRects; i++)
			myRects[i].render(someGraphics, myX, myY);

		myNumRects = 0;
	}

	//texts
	if(myNumTexts)
	{
		for(i = 0; i < myNumTexts; i++)
			myTexts[i].render(someGraphics, myX, myY);
		myNumTexts = 0;
	}
}

const POINT Window::clientCursorPos() const
{
	POINT cp(myState.cursorPos());

	cp.x -= myX;
	cp.y -= myY;

	return cp;
}

const void* Window::dropOutside() const
{
	return myState.dropOutside(myX, myY, myWidth, myHeight);
}

void Window::frame()
{
	//area
	if(rect(myWindows.hasFocus(*this) ? FOCUS : WINDOW, 0, 0, myWidth, myHeight, false) && myState.isDown(VK_LBUTTON))
		myWantsFocus = true;

	//title bar
	if(drag(this, 0, 0, myWidth, buttonHeight(), myTitle))
	{
		myWantsFocus = true;
		POINT cp(myState.dragPos());
		myX = cp.x;
		myY = cp.y;
	}
}

//private
const bool Window::rect(const Style& aStyle, const int aX, const int aY, const int aWidth, const int aHeight, const bool aSetInside)
{
	if(myNumRects < MAX_RECTS)
	{
		Rect& r(myRects[myNumRects]);
		myNumRects++;

		//only set down style if not dragging
		r.set(myState.isDown(VK_LBUTTON) && myState.myDrag == NULL, aStyle, aX, aY, aWidth, aHeight);

		//only set inside style if not dragging
		if(r.inside(myState.cursorPos(), myX, myY, aSetInside && myState.myDrag == NULL))
		{
			myState.myInsideRect = true;
			return true;
		}
	}

	return false;
}

const int Window::fromPercentX(const unsigned char aPercent) const
{
	return (int)((float)aPercent * 0.01f * (float)myWidth);
}

const int Window::fromPercentY(const unsigned char aPercent) const
{
	return (int)((float)aPercent * 0.01f * (float)myHeight);
}

bool& Window::handleState(const void* aHandle)
{
	static bool invalid(FALSE);
	unsigned int i;

	for(i = 0; i < myNumHandles; i++)
	{
		if(myHandles[i].myHandle == aHandle)
			return myHandles[i].myState;
	}

	if(myNumHandles < MAX_HANDLES)
	{
		myHandles[myNumHandles].myHandle = aHandle;
		return myHandles[myNumHandles++].myState;
	}
	else
	{
		return invalid;
	}
}

const bool Window::isDown(const int aKey) const
{
	return myWindows.hasFocus(*this) && myState.isDown(aKey);
}

const bool Window::wasPressed(const int aKey) const
{
	return myWindows.hasFocus(*this) && myState.wasPressed(aKey);
}

}